#include "call_graph.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "symlist.h"
#include "list.h"
#include "hist.h"

typedef struct {
  int caller;
  int called;
  int count;
} call;

static call *data;
static int size;
static int maxsize;

int add_call_item(unsigned long dig_addr1, unsigned long dig_addr2,
		  unsigned long count)
{
  int caller = find_symbol(dig_addr1);
  int called = find_symbol(dig_addr2);
  int i;

  if (caller == -1 || called == -1) {
    fprintf(stderr, "WARNING: unresolved caller (%8.8lx) /called (%8.8lx)\n",
	    dig_addr1, dig_addr2);
    return 0;
  }

  /* is it there? */
  /* barbarian - needs optimization (use a heap?) */
  for (i=0; i<size; i++) {
    if (data[i].caller == caller && data[i].called == called) {
      data[i].count += count;
      return 0;
    }
  }

  /* not there - add it */
  if (size == maxsize) {
    call *new;
    int ns = maxsize + 2048;
    if (data)
      new = realloc(data, ns * sizeof(call));
    else
      new = malloc(ns * sizeof(call));
    if (!new) return -1;
    data = new;
    maxsize = ns;
  }

  data[size].caller = caller;
  data[size].called = called;
  data[size].count = count;

  size++;

  return 0;
}

int add_call(int caller, int called, int count)
{
  if (list_add(&symbols[caller].call, called) == -1 ||
      list_add(&symbols[caller].call, count) == -1)
    return -1;
  return 0;
}

int add_called(int caller, int called, int count)
{
  if (list_add(&symbols[called].called_by, caller) == -1 ||
      list_add(&symbols[called].called_by, count) == -1)
    return -1;
  symbols[called].count += count;
  return 0;
}

int end_items(void)
{
  int i;
  int n = nb_symbols();

  symbols = malloc(n * sizeof(sym_info));
  if (!symbols) return -1;
  memset(symbols, 0, n * sizeof(sym_info));

  for (i = 0; i < size; i++) {
    if (add_call(data[i].caller, data[i].called, data[i].count) == -1 ||
        add_called(data[i].caller, data[i].called, data[i].count) == -1)
      return -1;
  }

  return 0;
}

static void dump_list(list *l)
{
  int i;
  int n = list_size(l);

  if (n == 0) {
    fprintf(stdout, "  <NONE>\n");
    return;
  }

  for (i = 0; i < n; i+=2) {
    fprintf(stdout, "  %s [%d] {%d} (%f | C1 %f | C2 %f) \n",
            symbol_name(list_get(l, i)),
            list_get(l, i+1),
            symbols[list_get(l, i)].count,
            symbols[list_get(l, i)].internal_time / 100,
            symbols[list_get(l, i)].cumuled_time_1 / 100,
            symbols[list_get(l, i)].cumuled_time_2 / 100);
  }
}

void dump_calls(void)
{
  int i;
  int n = nb_unique_symbols();
  sym_info *b;

  for (i = 0; i < n; i++) {
    //b = symbols + sorted_symbols[n-1-i];
    b = symbols + sorted_symbols_by_cumuled_2[n-1-i];
    if (!list_size(&b->call) && !list_size(&b->called_by) && !b->count)
      continue;
    fprintf(stdout, "---\n");
    fprintf(stdout, "%s [%d] (%f | C1 %f | C2 %f)\n",
      //symbol_name(sorted_symbols[n-1-i]), b->count,
      symbol_name(sorted_symbols_by_cumuled_2[n-1-i]), b->count,
      b->internal_time / 100,
      b->cumuled_time_1 / 100,
      b->cumuled_time_2 / 100);
    fprintf(stdout, "called by:\n");
    dump_list(&b->called_by);
    fprintf(stdout, "calls:\n");
    dump_list(&b->call);
  }
}
