/* a basic (and certainly buggy) C++ demangler
 * part of xgprof
 * Documentation for the mangling language:
 * http://www.codesourcery.com/cxx-abi/abi.html#mangling
 * May I say the documentation is very bad written? Well, I do.
 * See the end of this file of the grammar I use (in case there is
 * bug in there it might be good to check if the grammar I use is
 * up-to-date or correct) (but of course the bug might have nothing
 * to do with the grammar) (heck you get the point).
 * Public domain stuff. Copy it, adapt it, smoke it, own it, I don't
 * give the lesser shit.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "sstring.h"

#ifndef USE_CPP_DEMANGLER

typedef struct {
  char *data;
  int size;
  int maxsize;
} sbuf;

static int sput(sbuf *s, char c)
{
  if (s->size == s->maxsize) {
    char *new;
    int ns = s->maxsize + 512;
    if (s->maxsize)
      new = realloc(s->data, ns);
    else
      new = malloc(ns);
    if (!new) return 0;
    s->data = new;
    s->maxsize = ns;
  }

  s->data[s->size] = c;
  s->size++;

  return 1;
}

static int sputs(sbuf *s, char *b)
{
  while(*b) {
    if (!sput(s, *b)) return 0;
    b++;
  }
  return 1;
}

typedef struct {
  int *data;
  int size;
  int maxsize;
} subst;

int add_subst(subst *s, int pos)
{
  if (s->size == s->maxsize) {
    int *new;
    int ns = s->maxsize + 256;
    if (s->maxsize)
      new = realloc(s->data, ns * sizeof(int));
    else
      new = malloc(ns * sizeof(int));
    if (!new) return 0;
    s->data = new;
    s->maxsize = ns;
  }

  s->data[s->size] = pos;
  s->size++;

  return 1;
}

/* wonderful globals */
static char *source;
static int pos;

static sbuf names;
static sbuf out;
static subst subs;
static int templates;

#define TYPE_TEMPLATE_PARAM				1
#define NESTED_NAME					2
#define LOCAL_NAME					3
#define SOURCE_NAME					4
#define CTOR_DTOR_NAME					5
#define OPERATOR_NAME					6
#define SUBSTITUTION					7
#define EXPRESSION					8
#define EXPR_PRIMARY					9
#define TYPE						10
#define TEMPLATE_ARGS					11
#define NAME						12
#define CALL_OFFSET					13
#define SPECIAL_NAME					14
#define BARE_FUNCTION_TYPE				15
#define ENCODING					16

#define _EL(e, b)					elem_data[e].d[b]

#define ELEM_TYPE(e)					elem_data[e].type

#define TEMPLATE_PARAM_INDEX(e)				_EL(e, 0)

#define NESTED_NAME_NEXT(e)				_EL(e, 0)
#define NESTED_NAME_QUALIFIER_R(e)			_EL(e, 1)
#define NESTED_NAME_QUALIFIER_V(e)			_EL(e, 2)
#define NESTED_NAME_QUALIFIER_K(e)			_EL(e, 3)
#define NESTED_NAME_DATA(e)				_EL(e, 4)

#define LOCAL_NAME_ENCODING(e)				_EL(e, 0)
#define LOCAL_NAME_ENTITY(e)				_EL(e, 1)
#define LOCAL_NAME_DISCRIMINATOR(e)			_EL(e, 2)

#define SOURCE_NAME_STRING(e)				_EL(e, 0)

/* if CTOR_DTOR_NAME_CTOR(e) == 0 then it's a dtor */
#define CTOR_DTOR_NAME_CTOR(e)				_EL(e, 0)
#define CTOR_DTOR_NAME_DTOR(e)				_EL(e, 1)

#define OPERATOR_NAME_OP(e)				_EL(e, 0)
#define OPERATOR_NAME_NB_SONS(e)			_EL(e, 1)
#define OPERATOR_NAME_TYPE(e)				_EL(e, 2)
#define OPERATOR_NAME_SOURCE_NAME(e)			_EL(e, 3)
#define OPERATOR_NAME_SON(e, i)				_EL(e, 4+i)

#define SUBSTITUTION_TYPE(e)				_EL(e, 0)
#define SUBSTITUTION_N(e)				_EL(e, 1)

#define EXPRESSION_TYPE(e)				_EL(e, 0)
#define EXPRESSION_st_TYPE(e)				_EL(e, 1)
#define EXPRESSION_sr_TYPE(e)				_EL(e, 1)
#define EXPRESSION_sr_UNQUALIFIED_NAME(e)		_EL(e, 2)
#define EXPRESSION_sr_TEMPLATE_ARGS(e)			_EL(e, 3)
#define EXPRESSION_sT_TEMPLATE_PARAMS(e)		_EL(e, 1)
#define EXPRESSION_sL_EXPR_PRIMARY(e)			_EL(e, 1)
#define EXPRESSION_OPERATOR(e)				_EL(e, 1)

#define EXPR_PRIMARY_TYPE(e)				_EL(e, 0)
#define EXPR_PRIMARY_VALUE_TYPE(e)			_EL(e, 1)
#define EXPR_PRIMARY_VALUE_NUMBER(e)			_EL(e, 2)
#define EXPR_PRIMARY_MANGLED_NAME(e)			_EL(e, 1)

#define TYPE_TYPE(e)					_EL(e, 0)
#define TYPE_CLASS_ENUM_NAME(e)				_EL(e, 1)
#define TYPE_BUILTIN_TYPE(e)				_EL(e, 1)
#define TYPE_U2_SOURCE_NAME(e)				_EL(e, 1)
#define TYPE_F_FN_TYPE(e)				_EL(e, 1)
#define TYPE_A_TYPE(e)					_EL(e, 1)
#define TYPE_A_NUMBER(e)				_EL(e, 2)
#define TYPE_A_TYPE_TYPE(e)				_EL(e, 3)
#define TYPE_A_EXPR(e)					_EL(e, 2)
#define TYPE_M_TYPE1(e)					_EL(e, 1)
#define TYPE_M_TYPE2(e)					_EL(e, 2)
#define TYPE_T_TEMPLATE_PARAM(e)			_EL(e, 1)
#define TYPE_T_TEMPLATE_ARGS(e)				_EL(e, 2)
#define TYPE_S_SUBSTITUTION(e)				_EL(e, 1)
#define TYPE_S_TEMPLATE_ARGS(e)				_EL(e, 2)
#define TYPE_X_TYPE(e)					_EL(e, 1)
#define TYPE_X_SUBTYPE(e)				_EL(e, 2)
#define TYPE_U_SOURCE_NAME(e)				_EL(e, 1)
#define TYPE_U_TYPE(e)					_EL(e, 2)

#define TEMPLATE_ARGS_DATA(e)				_EL(e, 0)
#define TEMPLATE_ARGS_NEXT(e)				_EL(e, 1)

#define NAME_NAME(e)					_EL(e, 0)
#define NAME_TEMPLATE_ARGS(e)				_EL(e, 1)
#define NAME_ST(e)					_EL(e, 2)

#define CALL_OFFSET_NB(e)				_EL(e, 0)
#define CALL_OFFSET_1(e)				_EL(e, 1)
#define CALL_OFFSET_2(e)				_EL(e, 2)

#define SPECIAL_NAME_TYPE(e)				_EL(e, 0)
#define SPECIAL_NAME_NAME(e)				_EL(e, 1)
#define SPECIAL_NAME_TYPE_TYPE(e)			_EL(e, 1)
#define SPECIAL_NAME_ENCODING(e)			_EL(e, 1)
#define SPECIAL_NAME_CALL_OFFSET_1(e)			_EL(e, 2)
#define SPECIAL_NAME_CALL_OFFSET_2(e)			_EL(e, 3)

#define BARE_FUNCTION_TYPE_TYPE(e)			_EL(e, 0)
#define BARE_FUNCTION_TYPE_NEXT(e)			_EL(e, 1)

#define ENCODING_NAME(e)				_EL(e, 0)
#define ENCODING_BARE_FUNCTION_TYPE(e)			_EL(e, 1)

typedef struct {
  int type;
  int d[7];
} elem;

static elem *elem_data;
static int elem_size;
static int elem_maxsize;

int new_elem(void)
{
  if (elem_size == elem_maxsize) {
    elem *new;
    int ns = elem_maxsize + 256;
    if (elem_maxsize)
      new = realloc(elem_data, ns * sizeof(elem));
    else
      new = malloc(ns * sizeof(elem));
    if (!new) return 0;
    elem_data = new;
    elem_maxsize = ns;
    if (!elem_size) elem_size = 1; /* elem 0 is not used */
  }
  memset(elem_data + elem_size, 0, sizeof(elem));
  elem_size++;
  return elem_size-1;
}

static int unqualified_name(void);
static int template_args(void);
static int substitution(void);
static int mangled_name(void);
static int type(void);
static int bare_function_type(void);
static int name(void);
static int expr_primary(void);
static int encoding(void);

static int template_param(void)
{
  int sp = pos;
  int n=-1;
  int e = new_elem();
  if (!e) goto tp_err;

  if (source[pos] != 'T') goto tp_err;
  pos++;
  if (isdigit(source[pos])) {
    n = 0;
    do {
      n *= 10;
      n += source[pos] - '0';
      pos++;
    } while (isdigit(source[pos]));
  }
  n++;
  if (source[pos] != '_') goto tp_err;
  pos++;

  ELEM_TYPE(e) = TYPE_TEMPLATE_PARAM;
  TEMPLATE_PARAM_INDEX(e) = n;

  return e;

tp_err:
  pos = sp;
  return 0;
}

static int nested_name(void)
{
  int sp = pos;
  int e;
  int next;
  int first;

  if (source[pos] != 'N') goto nest_err;
  pos++;

  e = new_elem(); if (!e) goto nest_err;
  ELEM_TYPE(e) = NESTED_NAME;
  first = e;

  /* cv qualifiers */
  if (source[pos] == 'r') {
    NESTED_NAME_QUALIFIER_R(e) = 1;
    pos++;
  }
  if (source[pos] == 'V') {
    NESTED_NAME_QUALIFIER_V(e) = 1;
    pos++;
  }
  if (source[pos] == 'K') {
    NESTED_NAME_QUALIFIER_K(e) = 1;
    pos++;
  }

  while (1) {
    if (source[pos] == 'T') {
      if (!(NESTED_NAME_DATA(e)=template_param())) goto nest_err;
    } else if (source[pos] == 'S') {
      if (!(NESTED_NAME_DATA(e)=substitution())) goto nest_err;
    } else if (source[pos] == 'I') {
      if (!(NESTED_NAME_DATA(e)=template_args())) goto nest_err;
    } else {
      if (!(NESTED_NAME_DATA(e)=unqualified_name())) goto nest_err;
    }
    if (source[pos] == 'E') break;
    next = new_elem(); if (!next) goto nest_err;
    ELEM_TYPE(next) = NESTED_NAME;
    NESTED_NAME_NEXT(e) = next;
    e = next;
  }

  pos++;
  return first;

nest_err:
  pos = sp;
  return 0;
}

static int local_name(void)
{
  int sp = pos;
  int e = new_elem();
  if (!e) goto loc_err;

  ELEM_TYPE(e) = LOCAL_NAME;

  if (source[pos] != 'Z') goto loc_err;
  pos++;

  if (!(LOCAL_NAME_ENCODING(e)=encoding())) goto loc_err;

  if (source[pos] != 'E') goto loc_err;
  pos++;

  /* problem here 's' might be the start of st or sz operator
   * which should be handled by the else code. It is theoretically
   * possible, but practically? Anyway if there is a problem
   * at least this comment may help in debugging...
   */
  if (source[pos] == 's') {
    pos++;
  } else {
    if (!(LOCAL_NAME_ENTITY(e)=name())) goto loc_err;
  }

  if (source[pos] == '_') {
    int n = 0;
    pos++;
    if (!isdigit(source[pos])) goto loc_err;
    do {
      n *= 10;
      n += source[pos] - '0';
      pos++;
    } while (isdigit(source[pos]));
    LOCAL_NAME_DISCRIMINATOR(e) = n;
  }

  return e;

loc_err:
  pos = sp;
  return 0;
}

static int source_name(void)
{
  int n=0;
  int sp = pos;
  int e = new_elem();
  if (!e) return 0;
  ELEM_TYPE(e) = SOURCE_NAME;
  SOURCE_NAME_STRING(e) = names.size;

  if (!isdigit(source[pos])) goto sn_err;
  do {
    n *= 10;
    n += source[pos] - '0';
    pos++;
  } while (isdigit(source[pos]));
  while (n && source[pos]) {
    if (!sput(&names, source[pos])) goto sn_err;
    n--;
    pos++;
  }
  if (n) goto sn_err;
  if (!sput(&names, 0)) goto sn_err;
  return e;

sn_err:
  pos = sp;
  return 0;
}

static int ctor_dtor_name(void)
{
  int e = new_elem();
  if (!e) return 0;
  ELEM_TYPE(e) = CTOR_DTOR_NAME;

  if (source[pos] == 'C') {
    switch (source[pos+1]) {
      default: return 0;
      case '1': case '2': case '3':
        CTOR_DTOR_NAME_CTOR(e) = source[pos] - '0';
        pos+=2;
        return e;
    }
  }
  if (source[pos] == 'D') {
    switch (source[pos+1]) {
      default: return 0;
      case '0': case '1': case '2':
        CTOR_DTOR_NAME_DTOR(e) = source[pos] - '0';
        pos+=2;
        return e;
    }
  }
  return 0;
}

#define OP_UNARY_AND			1
#define OP_AND				2
#define OP_AFFECT			3
#define OP_AND_EQ			4
#define OP_AND_AND			5
#define OP_NEG				6
#define OP_COMMA			7
#define OP_CLASS			8
#define OP_CAST				9
#define OP_DELETE			10
#define OP_DELETE_ARRAY			11
#define OP_POINTER			12
#define OP_DIV				13
#define OP_DIV_EQ			14
#define OP_XOR				15
#define OP_XOR_EQ			16
#define OP_EQ				17
#define OP_GT				18
#define OP_GE				19
#define OP_ARRAY			20
#define OP_LEFT_SHIFT			21
#define OP_LEFT_SHIFT_EQ		22
#define OP_LT				23
#define OP_LE				24
#define OP_MINUS			25
#define OP_MUL				26
#define OP_MINUS_EQ			27
#define OP_MUL_EQ			28
#define OP_DEC				29
#define OP_NEW				30
#define OP_NEW_ARRAY			31
#define OP_MINUS_UNARY			32
#define OP_NEQ				33
#define OP_NOT				34
#define OP_OR				35
#define OP_OR_EQ			36
#define OP_OR_OR			37
#define OP_PLUS_UNARY			38
#define OP_PLUS				39
#define OP_PLUS_EQ			40
#define OP_PLUS_PLUS			41
#define OP_PM				42
#define OP_PT				43
#define OP_QUESTION			44
#define OP_MOD				45
#define OP_MOD_EQ			46
#define OP_RIGHT_SHIFT			47
#define OP_RIGHT_SHIFT_EQ		48
#define OP_SIZEOF			49
#define OP_VENDOR			50

static int operator_name(int *nb_params)
{
  char a, b;
  int e = new_elem();
  if (!e) return 0;
  ELEM_TYPE(e) = OPERATOR_NAME;

  a = source[pos]; if (!a) return 0;
  b = source[pos+1]; if (!b) return 0;

  switch (a) {
    default: return 0;
    case 'a': switch (b) {
        default: return 0;
        case 'd':
          OPERATOR_NAME_OP(e) = OP_UNARY_AND;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          break;
        case 'n':
          OPERATOR_NAME_OP(e) = OP_AND;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'S':
          OPERATOR_NAME_OP(e) = OP_AFFECT;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'N':
          OPERATOR_NAME_OP(e) = OP_AND_EQ;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'a':
          OPERATOR_NAME_OP(e) = OP_AND_AND;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
      }
      break;
    case 'c': switch (b) {
        default: return 0;
        case 'o':
          OPERATOR_NAME_OP(e) = OP_NEG;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          break;
        case 'm':
          OPERATOR_NAME_OP(e) = OP_COMMA;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'l':
          OPERATOR_NAME_OP(e) = OP_CLASS;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          break;
        case 'v': {
          int sp = pos;
          pos += 2;
          if (!(OPERATOR_NAME_TYPE(e)=type())) { pos = sp; return 0; }
          OPERATOR_NAME_OP(e) = OP_CAST;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          return e;
        }
      }
      break;
    case 'd': switch (b) {
        default: return 0;
        case 'l':
          OPERATOR_NAME_OP(e) = OP_DELETE;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          break;
        case 'a':
          OPERATOR_NAME_OP(e) = OP_DELETE_ARRAY;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          break;
        case 'e':
          OPERATOR_NAME_OP(e) = OP_POINTER;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          break;
        case 'v':
          OPERATOR_NAME_OP(e) = OP_DIV;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'V':
          OPERATOR_NAME_OP(e) = OP_DIV_EQ;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
      }
      break;
    case 'e': switch (b) {
        default: return 0;
        case 'o':
          OPERATOR_NAME_OP(e) = OP_XOR;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'O':
          OPERATOR_NAME_OP(e) = OP_XOR_EQ;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'q':
          OPERATOR_NAME_OP(e) = OP_EQ;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
      }
      break;
    case 'g': switch (b) {
        default: return 0;
        case 't':
          OPERATOR_NAME_OP(e) = OP_GT;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'e':
          OPERATOR_NAME_OP(e) = OP_GE;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
      }
      break;
    case 'i': switch (b) {
        default: return 0;
        case 'x':
          OPERATOR_NAME_OP(e) = OP_ARRAY;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
      }
      break;
    case 'l': switch (b) {
        default: return 0;
        case 's':
          OPERATOR_NAME_OP(e) = OP_LEFT_SHIFT;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'S':
          OPERATOR_NAME_OP(e) = OP_LEFT_SHIFT_EQ;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 't':
          OPERATOR_NAME_OP(e) = OP_LT;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'e':
          OPERATOR_NAME_OP(e) = OP_LE;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
      }
      break;
    case 'm': switch (b) {
        default: return 0;
        case 'i':
          OPERATOR_NAME_OP(e) = OP_MINUS;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'l':
          OPERATOR_NAME_OP(e) = OP_MUL;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'I':
          OPERATOR_NAME_OP(e) = OP_MINUS_EQ;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'L':
          OPERATOR_NAME_OP(e) = OP_MUL_EQ;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'm':
          OPERATOR_NAME_OP(e) = OP_DEC;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          break;
      }
      break;
    case 'n': switch (b) {
        default: return 0;
        case 'w':
          OPERATOR_NAME_OP(e) = OP_NEW;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          break;
        case 'a':
          OPERATOR_NAME_OP(e) = OP_NEW_ARRAY;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          break;
        case 'g':
          OPERATOR_NAME_OP(e) = OP_MINUS_UNARY;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          break;
        case 'e':
          OPERATOR_NAME_OP(e) = OP_NEQ;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 't':
          OPERATOR_NAME_OP(e) = OP_NOT;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          break;
      }
      break;
    case 'o': switch (b) {
        default: return 0;
        case 'r':
          OPERATOR_NAME_OP(e) = OP_OR;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'R':
          OPERATOR_NAME_OP(e) = OP_OR_EQ;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'o':
          OPERATOR_NAME_OP(e) = OP_OR_OR;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
      }
      break;
    case 'p': switch (b) {
        default: return 0;
        case 's':
          OPERATOR_NAME_OP(e) = OP_PLUS_UNARY;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          break;
        case 'l':
          OPERATOR_NAME_OP(e) = OP_PLUS;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'L':
          OPERATOR_NAME_OP(e) = OP_PLUS_EQ;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'p':
          OPERATOR_NAME_OP(e) = OP_PLUS_PLUS;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          break;
        case 'm':
          OPERATOR_NAME_OP(e) = OP_PM;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 't':
          OPERATOR_NAME_OP(e) = OP_PT;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
      }
      break;
    case 'q': switch (b) {
        default: return 0;
        case 'u':
          OPERATOR_NAME_OP(e) = OP_QUESTION;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 3;
          break;
      }
      break;
    case 'r': switch (b) {
        default: return 0;
        case 'm':
          OPERATOR_NAME_OP(e) = OP_MOD;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'M':
          OPERATOR_NAME_OP(e) = OP_MOD_EQ;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 's':
          OPERATOR_NAME_OP(e) = OP_RIGHT_SHIFT;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
        case 'S':
          OPERATOR_NAME_OP(e) = OP_RIGHT_SHIFT_EQ;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 2;
          break;
      }
      break;
    case 's': switch (b) {
        default: return 0;
        case 't':
        case 'z':
          OPERATOR_NAME_OP(e) = OP_SIZEOF;
          OPERATOR_NAME_NB_SONS(e) = *nb_params = 1;
          break;
      }
      break;
    case 'v': {
      int sp = pos;
      if (!isdigit(b)) return 0;
      pos+=2;
      if (!(OPERATOR_NAME_SOURCE_NAME(e)=source_name())) { pos=sp; return 0; }
      OPERATOR_NAME_OP(e) = OP_VENDOR;
      OPERATOR_NAME_NB_SONS(e) = *nb_params = b - '0';
      if (*nb_params < 1 || *nb_params > 3) { pos = sp; return 0; }
      return e;
    }
  }

  pos+=2;
  return e;
}

static int unqualified_name(void)
{
  if (isdigit(source[pos]))
    return source_name();
  else if (source[pos] == 'C' || source[pos] == 'D')
    return ctor_dtor_name();
  else {
    int dummy;
    return operator_name(&dummy);
  }
}

static int substitution(void)
{
  int sp = pos;
  int e = new_elem();
  if (!e) goto subst_err;
  ELEM_TYPE(e) = SUBSTITUTION;

  if (source[pos] != 'S') goto subst_err;
  pos++;

  if (source[pos] == '_') { pos++; return e; }
  if (isdigit(source[pos]) || (source[pos] >= 'A' && source[pos] <= 'Z')) {
    int n = 0;
    do {
      n *= 36;
      if (isdigit(source[pos]))
        n += source[pos] - '0';
      else
        n += 10 + source[pos] - 'A';
      pos++;
    } while (isdigit(source[pos]) ||
             (source[pos] >= 'A' && source[pos] <= 'Z'));
    if (source[pos] != '_') goto subst_err;
    SUBSTITUTION_N(e) = n+1;
    pos++;
    return e;
  }

  switch (source[pos]) {
    default: goto subst_err;
    case 't': case 'a': case 'b': case 's': case 'i': case 'o': case 'd':
      SUBSTITUTION_TYPE(e) = source[pos];
      pos++;
  }
  return e;

subst_err:
  pos = sp;
  return 0;
}

#define EXP_ST	1
#define EXP_SR	2
#define EXP_L	3
#define EXP_OP	4
#define EXP_T	5

static int expression(void)
{
  int nb_exp;
  int i;
  int sp = pos;
  int e = new_elem();
  if (!e) goto exp_err;

  ELEM_TYPE(e) = EXPRESSION;

  if (source[pos] == 's' && source[pos+1] == 't') {
    pos+=2;
    if (!(EXPRESSION_st_TYPE(e)=type())) goto exp_err;
    EXPRESSION_TYPE(e) = EXP_ST;
    return e;
  }

  if (source[pos] == 's' && source[pos+1] == 'r') {
    pos+=2;
    EXPRESSION_TYPE(e) = EXP_SR;
    if (!(EXPRESSION_sr_TYPE(e)=type())) goto exp_err;
    if (!(EXPRESSION_sr_UNQUALIFIED_NAME(e)=unqualified_name())) goto exp_err;
    if (source[pos] != 'I') return e;
    if (!(EXPRESSION_sr_TEMPLATE_ARGS(e)=template_args())) goto exp_err;
    return e;
  }

  if (source[pos] == 'T') {
    EXPRESSION_TYPE(e) = EXP_T;
    if (!(EXPRESSION_sT_TEMPLATE_PARAMS(e)=template_param())) goto exp_err;
    return e;
  }

  if (source[pos] == 'L') {
    EXPRESSION_TYPE(e) = EXP_L;
    if (!(EXPRESSION_sL_EXPR_PRIMARY(e)=expr_primary())) goto exp_err;
    return e;
  }

  EXPRESSION_TYPE(e) = EXP_OP;
  if (!(EXPRESSION_OPERATOR(e)=operator_name(&nb_exp))) goto exp_err;

  i = 0;
  while (nb_exp) {
    if (!(OPERATOR_NAME_SON(EXPRESSION_OPERATOR(e), i)=expression()))
      goto exp_err;
    nb_exp--;
    i++;
  }

  return e;

exp_err:
  pos = sp;
  return 0;
}

#define PRIMEXP_NAME	1
#define PRIMEXP_VALUE	2

static int expr_primary(void)
{
  int sp = pos;
  int e = new_elem();
  if (!e) goto ep_err;

  ELEM_TYPE(e) = EXPR_PRIMARY;

  if (source[pos] != 'L') goto ep_err;
  pos++;

  if (source[pos] == '_') {
    EXPR_PRIMARY_TYPE(e) = PRIMEXP_NAME;
    if (!(EXPR_PRIMARY_MANGLED_NAME(e)=mangled_name())) goto ep_err;
  } else {
    EXPR_PRIMARY_TYPE(e) = PRIMEXP_VALUE;
    if (!(EXPR_PRIMARY_VALUE_TYPE(e)=type())) goto ep_err;
    /* number of float */
    EXPR_PRIMARY_VALUE_NUMBER(e) = names.size;
    if (source[pos] == 'n') {
      if (!sput(&names, '-')) goto ep_err;
      pos++; /* negative number */
    }
    while (source[pos] && source[pos] != 'E') {
      if (!sput(&names, source[pos])) goto ep_err;
      pos++;
    }
    if (!sput(&names, 0)) goto ep_err;
  }

  if (source[pos] != 'E') goto ep_err;
  pos++;

  return e;

ep_err:
  pos = sp;
  return 0;
}

#define TP_CLASS_ENUM	1
#define TP_BUILTIN	2
#define TP_U2		3
#define TP_F		4
#define TP_A		5
#define TP_M		6
#define TP_T		7
#define TP_S		8
#define TP_X		9
#define TP_U		10

#define TP_A_EXPR	1
#define TP_A_NUMBER	2

static int type(void)
{
  int sp = pos;
  int e = new_elem();
  if (!e) goto type_err;

  ELEM_TYPE(e) = TYPE;

  switch (source[pos]) {
    default:
      /* class-enum-type maybe */
type_enum:
      TYPE_TYPE(e) = TP_CLASS_ENUM;
      if (!(TYPE_CLASS_ENUM_NAME(e)=name())) goto type_err;
      break;
    case 'v':
    case 'w':
    case 'b':
    case 'c':
    case 'a':
    case 'h':
    case 's':
    case 't':
    case 'i':
    case 'j':
    case 'l':
    case 'm':
    case 'x':
    case 'y':
    case 'n':
    case 'o':
    case 'f':
    case 'd':
    case 'e':
    case 'g':
    case 'z':
      TYPE_TYPE(e) = TP_BUILTIN;
      TYPE_BUILTIN_TYPE(e) = source[pos];
      pos++;
      break;
    case 'u':
      TYPE_TYPE(e) = TP_U2;
      pos++;
      if (!(TYPE_U2_SOURCE_NAME(e)=source_name())) goto type_err;
      break;
    case 'F':
      TYPE_TYPE(e) = TP_F;
      pos++;
      if (source[pos] == 'Y') pos++;
      if (!(TYPE_F_FN_TYPE(e)=bare_function_type())) goto type_err;
      if (source[pos] != 'E') goto type_err;
      pos++;
      break;
    case 'A':
      TYPE_TYPE(e) = TP_A;
      pos++;
      if (isdigit(source[pos])) {
        int n=0;
        TYPE_A_TYPE(e) = TP_A_NUMBER;
        do {
          n *= 10;
          n += source[pos] - '0';
          pos++;
        } while (source[pos] && isdigit(source[pos]));
        TYPE_A_NUMBER(e) = n;
        if (!source[pos]) goto type_err;
        if (source[pos] != '_') goto type_err;
        pos++;
      } else {
        TYPE_A_TYPE(e) = TP_A_EXPR;
        if (source[pos] != '_') {
          if (!(TYPE_A_EXPR(e)=expression())) goto type_err;
          if (source[pos] != '_') goto type_err;
        }
        pos++;
      }
      if (!(TYPE_A_TYPE_TYPE(e)=type())) goto type_err;
      break;
    case 'M':
      TYPE_TYPE(e) = TP_M;
      pos++;
      if (!(TYPE_M_TYPE1(e)=type())) goto type_err;
      if (!(TYPE_M_TYPE2(e)=type())) goto type_err;
      break;
    case 'T':
      TYPE_TYPE(e) = TP_T;
      if (!(TYPE_T_TEMPLATE_PARAM(e)=template_param())) goto type_err;
      if (source[pos] == 'I') {
        if (!(TYPE_T_TEMPLATE_ARGS(e)=template_args())) goto type_err;
      }
      break;
    case 'S':
      /* if we get St it should be a prefix for a name */
      if (source[pos+1] == 't') goto type_enum;
      TYPE_TYPE(e) = TP_S;
      if (!(TYPE_S_SUBSTITUTION(e)=substitution())) goto type_err;
      if (source[pos] == 'I') {
        if (!(TYPE_S_TEMPLATE_ARGS(e)=template_args())) goto type_err;
      }
      break;
    case 'r':
    case 'V':
    case 'K':
      TYPE_TYPE(e) = TP_X;
      TYPE_X_SUBTYPE(e) = source[pos];
      pos++;
      if (!(TYPE_X_TYPE(e)=type())) goto type_err;
      break;
    case 'P':
    case 'R':
    case 'O':
    case 'C':
    case 'G':
      TYPE_TYPE(e) = TP_X;
      TYPE_X_SUBTYPE(e) = source[pos];
      pos++;
      if (!(TYPE_X_TYPE(e)=type())) goto type_err;
      break;
    case 'U':
      TYPE_TYPE(e) = TP_U;
      pos++;
      if (!(TYPE_U_SOURCE_NAME(e)=source_name())) goto type_err;
      if (!(TYPE_U_TYPE(e)=type())) goto type_err;
      break;
  }

  return e;

type_err:
  pos = sp;
  return 0;
}

static int template_args(void)
{
  int sp = pos;
  int e = new_elem();
  int next;
  int first;
  if (!e) goto ta_err;

  ELEM_TYPE(e) = TEMPLATE_ARGS;
  first = e;

  if (source[pos] != 'I') goto ta_err;
  pos++;

  while (1) {
    if (source[pos] == 'X') {
      pos++;
      if (!(TEMPLATE_ARGS_DATA(e)=expression())) goto ta_err;
      if (source[pos] != 'E') goto ta_err;
      pos++;
    } else if (source[pos] == 'L') {
      if (!(TEMPLATE_ARGS_DATA(e)=expr_primary())) goto ta_err;
    } else {
      if (!(TEMPLATE_ARGS_DATA(e)=type())) goto ta_err;
    }
    if (source[pos] == 'E') break;
    next = new_elem(); if (!next) goto ta_err;
    ELEM_TYPE(next) = TEMPLATE_ARGS;
    TEMPLATE_ARGS_NEXT(e) = next;
    e = next;
  }

  pos++;
  return first;

ta_err:
  pos = sp;
  return 0;
}

static int name(void)
{
  int e;

  if (source[pos] == 'N') return nested_name();
  if (source[pos] == 'Z') return local_name();

  e = new_elem(); if (!e) return 0;
  ELEM_TYPE(e) = NAME;

  if (source[pos] == 'S' && source[pos+1] == 't') {
    NAME_ST(e) = 1;
    pos += 2;
  }

  /* unscoped template name? */
  if (source[pos] == 'S') {
    if (!(NAME_NAME(e)=substitution())) return 0;
  } else {
    if (!(NAME_NAME(e)=unqualified_name())) return 0;
  }

  if (source[pos] == 'I') {
    if (!(NAME_TEMPLATE_ARGS(e)=template_args())) return 0;
  }

  return e;
}

static int call_offset(void)
{
  int n;
  int sign;
  int e = new_elem();
  if (!e) return 0;

  ELEM_TYPE(e) = CALL_OFFSET;

  if (source[pos] == 'h') {
    CALL_OFFSET_NB(e) = 1;
    pos++;
    if (source[pos] == 'n') {
      sign = -1;
      pos++;
    } else sign = 1;
    n = 0;
    while (isdigit(source[pos])) {
      n *= 10;
      n += source[pos] - '0';
      pos++;
    }
    if (source[pos] != '_') return 0;
    pos++;
    CALL_OFFSET_1(e) = n * sign;
    return e;
  }

  if (source[pos] == 'v') {
    CALL_OFFSET_NB(e) = 2;
    pos++;
    if (source[pos] == 'n') {
      sign = -1;
      pos++;
    } else sign = 1;
    n = 0;
    while (isdigit(source[pos])) {
      n *= 10;
      n += source[pos] - '0';
      pos++;
    }
    if (source[pos] != '_') return 0;
    pos++;
    CALL_OFFSET_1(e) = n * sign;
    if (source[pos] == 'n') {
      sign = -1;
      pos++;
    } else sign = 1;
    n = 0;
    while (isdigit(source[pos])) {
      n *= 10;
      n += source[pos] - '0';
      pos++;
    }
    if (source[pos] != '_') return 0;
    pos++;
    CALL_OFFSET_2(e) = n * sign;
    return e;
  }

  return 0;
}

static int special_name(void)
{
  int sp = pos;
  int e = new_elem();
  if (!e) goto sn_err;
  ELEM_TYPE(e) = SPECIAL_NAME;

  if (source[pos] == 'G' && source[pos+1] == 'V') {
    SPECIAL_NAME_TYPE(e) = 'G';
    pos += 2;
    if (!(SPECIAL_NAME_NAME(e)=name())) goto sn_err;
    return e;
  }

  if (source[pos] != 'T') goto sn_err;
  pos++;

  switch (source[pos]) {
    case 'V':
    case 'T':
    case 'I':
    case 'S':
      SPECIAL_NAME_TYPE(e) = source[pos];
      pos ++;
      if (!(SPECIAL_NAME_TYPE_TYPE(e)=type())) goto sn_err;
      break;
    case 'h':
    case 'v':
      /* no pos++ here */
      SPECIAL_NAME_TYPE(e) = source[pos];
      if (!(SPECIAL_NAME_CALL_OFFSET_1(e)=call_offset())) goto sn_err;
      if (!(SPECIAL_NAME_ENCODING(e)=encoding())) goto sn_err;
      break;
    case 'c':
      SPECIAL_NAME_TYPE(e) = source[pos];
      pos ++;
      if (!(SPECIAL_NAME_CALL_OFFSET_1(e)=call_offset())) goto sn_err;
      if (!(SPECIAL_NAME_CALL_OFFSET_2(e)=call_offset())) goto sn_err;
      if (!(SPECIAL_NAME_ENCODING(e)=encoding())) goto sn_err;
      break;
  }

  return e;

sn_err:
  pos = sp;
  return 0;
}

static int bare_function_type(void)
{
  int e = new_elem();
  int next;
  int first;

  if (!e) return 0;
  ELEM_TYPE(e) = BARE_FUNCTION_TYPE;
  first = e;

  /* if you read the grammar, you see that after a bare-function-type
   * you may have end-of-string or 'E' and nothing else
   */
  do {
    if (!(BARE_FUNCTION_TYPE_TYPE(e)=type())) return 0;
    if (source[pos] == 0 || source[pos] == 'E') break;
    next = new_elem(); if (!next) return 0;
    ELEM_TYPE(next) = BARE_FUNCTION_TYPE;
    BARE_FUNCTION_TYPE_NEXT(e) = next;
    e = next;
  } while (1);

  return first;
}

static int encoding(void)
{
  int e = new_elem();
  if (!e) return 0;
  ELEM_TYPE(e) = ENCODING;

  if (source[pos] == 'T' || source[pos] == 'G') {
    ENCODING_NAME(e)=special_name();
    return e;
  }
  if (!(ENCODING_NAME(e)=name())) return 0;
  if (!source[pos] || source[pos] == 'E') return e;
  if (!(ENCODING_BARE_FUNCTION_TYPE(e)=bare_function_type())) return 0;
  return e;
}

static int mangled_name(void)
{
  if (source[pos] != '_' || source[pos+1] != 'Z') return 0;
  pos += 2;
  return encoding();
}

static int dump_unqualified_name(int);
static int dump_source_name(int);
static int dump_expression(int);
static int dump_expr_primary(int);
static int dump_type(int);
static int dump_ctor_dtor_name(int, int);
static int dump_operator_name(int, int);
static int dump_encoding(int);

static int dump_string(char *s)
{
  return sputs(&out, s);
}

static int dump_substitution(int e)
{
  char s[32];
  char *t;

  switch (SUBSTITUTION_TYPE(e)) {
    default:return 0;
    case 0:
      /* TODO */
      sprintf(s, "SUBST(%d)", SUBSTITUTION_N(e));
      t = s;
      break;
    case 't':
      t = "std";
      break;
    case 'a':
      t = "std::allocator";
      break;
    case 'b':
      t = "std::basic_string";
      break;
    case 's':
      t =
   "std::basic_string<char, std::char_traits<char>, std::allocator<char> >";

      break;
    case 'i':
      t = "std::basic_istream<char, std::char_traits<char> >";
      break;
    case 'o':
      t = "std::basic_ostream<char, std::char_traits<char> >";
      break;
    case 'd':
      t = "std::basic_iostream<char, std::char_traits<char> >";
      break;
  }

  if (!dump_string(t)) return 0;

  return 1;
}

static int dump_template_param(int e)
{
  int t = templates;
  int n = TEMPLATE_PARAM_INDEX(e);
  while (t && n) {
    t = TEMPLATE_ARGS_NEXT(t);
    n--;
  }
  if (!t) return 0;
  switch (ELEM_TYPE(TEMPLATE_ARGS_DATA(t))) {
    case EXPRESSION:
      if (!dump_expression(TEMPLATE_ARGS_DATA(t))) return 0;
      break;
    case EXPR_PRIMARY:
      if (!dump_expr_primary(TEMPLATE_ARGS_DATA(t))) return 0;
      break;
    case TYPE:
      if (!dump_type(TEMPLATE_ARGS_DATA(t))) return 0;
      break;
  }
  return 1;
}

static int dump_expr_primary(int e)
{
  switch (EXPR_PRIMARY_TYPE(e)) {
    case PRIMEXP_VALUE:
      if (!dump_string("(")) return 0;
      if (!dump_type(EXPR_PRIMARY_VALUE_TYPE(e))) return 0;
      if (!dump_string(")")) return 0;
      if (!dump_string(names.data + EXPR_PRIMARY_VALUE_NUMBER(e))) return 0;
      break;
    case PRIMEXP_NAME:
      if (!dump_encoding(EXPR_PRIMARY_MANGLED_NAME(e))) return 0;
      break;
  }

  return 1;
}

static int dump_template_args(int e)
{
  int cur = e;
  int next = TEMPLATE_ARGS_NEXT(e);

  if (!dump_string("<")) return 0;

  while (cur) {
    int f = TEMPLATE_ARGS_DATA(cur);
    switch (ELEM_TYPE(f)) {
      default: return 0;
      case EXPRESSION: if (!dump_expression(f)) return 0; break;
      case EXPR_PRIMARY: if (!dump_expr_primary(f)) return 0; break;
      case TYPE: if (!dump_type(f)) return 0; break;
    }
    if (next) if (!dump_string(", ")) return 0;
    cur = next;
    if (next) next = TEMPLATE_ARGS_NEXT(next);
  }

  if (out.data[out.size-1] == '>') if (!dump_string(" ")) return 0;
  if (!dump_string(">")) return 0;

  return 1;
}

static int dump_expression(int e)
{
  switch (EXPRESSION_TYPE(e)) {
    case EXP_ST:
      if (!dump_string("sizeof(")) return 0;
      if (!dump_type(EXPRESSION_st_TYPE(e))) return 0;
      if (!dump_string(")")) return 0;
      break;
    case EXP_SR:
      if (!dump_string("(")) return 0;
      if (!dump_type(EXPRESSION_sr_TYPE(e))) return 0;
      if (!dump_string(")(")) return 0;
      if (!dump_unqualified_name(EXPRESSION_sr_UNQUALIFIED_NAME(e))) return 0;
      if (EXPRESSION_sr_TEMPLATE_ARGS(e)) {
        if (!dump_string("<")) return 0;
        if (!dump_template_args(EXPRESSION_sr_TEMPLATE_ARGS(e))) return 0;
        if (out.data[out.size-1] == '>') if (!dump_string(" ")) return 0;
        if (!dump_string(">")) return 0;
      }
      if (!dump_string(")")) return 0;
      break;
    case EXP_L:
      if (!dump_expr_primary(EXPRESSION_sL_EXPR_PRIMARY(e))) return 0;
      break;
    case EXP_T:
      if (!dump_template_param(EXPRESSION_sT_TEMPLATE_PARAMS(e))) return 0;
      break;
    case EXP_OP: {
      int op = EXPRESSION_OPERATOR(e);
      if (!dump_string("(")) return 0;
      if (!dump_operator_name(op, 0)) return 0;
      if (!dump_string(" ")) return 0;
      switch (OPERATOR_NAME_NB_SONS(op)) {
        default: return 0;
        case 1:
          if (!dump_expression(OPERATOR_NAME_SON(op, 0))) return 0;
          break;
        case 2:
          if (!dump_expression(OPERATOR_NAME_SON(op, 0))) return 0;
          if (!dump_string(" ")) return 0;
          if (!dump_expression(OPERATOR_NAME_SON(op, 1))) return 0;
          break;
        case 3:
          if (!dump_expression(OPERATOR_NAME_SON(op, 0))) return 0;
          if (!dump_string(" ")) return 0;
          if (!dump_expression(OPERATOR_NAME_SON(op, 1))) return 0;
          if (!dump_string(" ")) return 0;
          if (!dump_expression(OPERATOR_NAME_SON(op, 2))) return 0;
          break;
      }
      if (!dump_string(")")) return 0;
      break;
    }
  }

  return 1;
}

static int dump_name(int e)
{
  if (ELEM_TYPE(e) == NESTED_NAME) {
    int next;
    int cur;
    cur = e;
    next = NESTED_NAME_NEXT(e);
    while (cur) {
      int el = NESTED_NAME_DATA(cur);
      switch (ELEM_TYPE(el)) {
        default: return 0;
        case TYPE_TEMPLATE_PARAM:
          if (!dump_template_param(el)) return 0;
          break;
        case SUBSTITUTION:
          if (!dump_substitution(el)) return 0;
          break;
        case TEMPLATE_ARGS:
          if (!dump_template_args(el)) return 0;
          break;
        case SOURCE_NAME:
          if (!dump_source_name(el)) return 0;
          break;
        case CTOR_DTOR_NAME:
          if (!dump_ctor_dtor_name(el, e)) return 0;
          break;
        case OPERATOR_NAME:
          if (!dump_operator_name(el, 1)) return 0;
          break;
      }
      if (next)
        if (ELEM_TYPE(NESTED_NAME_DATA(next)) != TEMPLATE_ARGS)
          if (!dump_string("::")) return 0;
      cur = next;
      if (next) next = NESTED_NAME_NEXT(cur);
    }
    if (NESTED_NAME_QUALIFIER_R(e))
      if (!dump_string(" restrict")) return 0;
    if (NESTED_NAME_QUALIFIER_V(e))
      if (!dump_string(" volatile")) return 0;
    if (NESTED_NAME_QUALIFIER_K(e))
      if (!dump_string(" const")) return 0;
    return 1;
  }

  if (ELEM_TYPE(e) == LOCAL_NAME) {
    if (!dump_encoding(LOCAL_NAME_ENCODING(e))) return 0;
    if (!dump_string("::")) return 0;
    if (LOCAL_NAME_ENTITY(e)) {
      if (!dump_name(LOCAL_NAME_ENTITY(e))) return 0;
    } else {
      if (!dump_string("string literal")) return 0;
    }
    return 1;
  }

  if (ELEM_TYPE(e) == SPECIAL_NAME) {
    /* <special-name> only appears the <encoding> and thus cannot
     * be seen here (it is handled in 'dump_encoding')
     */
    fprintf(stderr,
      "DEMANGLER: normally you should not read that, contact sed@free.fr\n");
    return 0;
  }

  if (NAME_ST(e)) if (!dump_string("std::")) return 0;

  if (ELEM_TYPE(NAME_NAME(e)) == SUBSTITUTION) {
    if (!dump_substitution(NAME_NAME(e))) return 0;
  } else {
    if (!dump_unqualified_name(NAME_NAME(e))) return 0;
  }

  if (NAME_TEMPLATE_ARGS(e))
    if (!dump_template_args(NAME_TEMPLATE_ARGS(e))) return 0;

  return 1;
}

static int dump_type(int e)
{
  switch (TYPE_TYPE(e)) {
    case TP_CLASS_ENUM:
      if (!dump_name(TYPE_CLASS_ENUM_NAME(e))) return 0;
      break;
    case TP_BUILTIN: {
      char *t = "UNKNOWN";
      switch (TYPE_BUILTIN_TYPE(e)) {
        case 'v': t = "void"; break;
        case 'w': t = "wchar_t"; break;
        case 'b': t = "bool"; break;
        case 'c': t = "char"; break;
        case 'a': t = "signed char"; break;
        case 'h': t = "unsigned char"; break;
        case 's': t = "short"; break;
        case 't': t = "unsigned short"; break;
        case 'i': t = "int"; break;
        case 'j': t = "unsigned int"; break;
        case 'l': t = "long"; break;
        case 'm': t = "unsigned long"; break;
        case 'x': t = "long long"; break;
        case 'y': t = "unsigned long long"; break;
        case 'n': t = "__int128"; break;
        case 'o': t = "unsigned __int128"; break;
        case 'f': t = "float"; break;
        case 'd': t = "double"; break;
        case 'e': t = "long double"; break;
        case 'g': t = "__float128"; break;
        case 'z': t = "..."; break;
      }
      if (!dump_string(t)) return 0;
      break;
    }
    case TP_U2:
      if (!dump_source_name(TYPE_U2_SOURCE_NAME(e))) return 0;
      break;
    case TP_F: {
      int f = TYPE_F_FN_TYPE(e);
      if (!dump_type(BARE_FUNCTION_TYPE_TYPE(f))) return 0;
      f = BARE_FUNCTION_TYPE_NEXT(f);
      if (!f) return 0;
      if (!dump_string(" ()(")) return 0;
      while (f) {
        if (!dump_type(BARE_FUNCTION_TYPE_TYPE(f))) return 0;
        f = BARE_FUNCTION_TYPE_NEXT(f);
        if (f) if (!dump_string(", ")) return 0;
      }
      if (!dump_string(")")) return 0;
      break;
    }
    case TP_A:
      /* TODO: don't do string->int->string for the dimension */
      if (!dump_type(TYPE_A_TYPE_TYPE(e))) return 0;
      if (!dump_string(" [")) return 0;
      if (TYPE_A_TYPE(e) == TP_A_NUMBER) {
        char b[32];
        sprintf(b, "%d", TYPE_A_NUMBER(e));
        if (!dump_string(b)) return 0;
      } else {
        if (TYPE_A_EXPR(e)) if (!dump_expression(TYPE_A_EXPR(e))) return 0;
      }
      if (!dump_string("]")) return 0;
      break;
    case TP_M:
      if (!dump_string("*(base class: ")) return 0;
      if (!dump_type(TYPE_M_TYPE1(e))) return 0;
      if (!dump_string(", member type: ")) return 0;
      if (!dump_type(TYPE_M_TYPE2(e))) return 0;
      if (!dump_string(")")) return 0;
      break;
    case TP_T:
      if (!dump_template_param(TYPE_T_TEMPLATE_PARAM(e))) return 0;
      if (TYPE_T_TEMPLATE_ARGS(e))
        if (!dump_template_args(TYPE_T_TEMPLATE_ARGS(e))) return 0;
      break;
    case TP_S:
      if (!dump_substitution(TYPE_S_SUBSTITUTION(e))) return 0;
      if (TYPE_S_TEMPLATE_ARGS(e))
        if (!dump_template_args(TYPE_S_TEMPLATE_ARGS(e))) return 0;
      break;
    case TP_X: {
      char *t = "UNKNOWN";
      switch (TYPE_X_SUBTYPE(e)) {
        case 'r': t = "restrict"; break;
        case 'V': t = "volatile"; break;
        case 'K': t = "const"; break;
        case 'P': t = "*"; break;
        case 'R': t = "&"; break;
        case 'O': t = "rvalue&"; break;
        case 'C': t = "complex pair"; break;
        case 'G': t = "imaginary"; break;
      }
      if (!dump_string(t) || !dump_string("{")) return 0;
      if (!dump_type(TYPE_X_TYPE(e))) return 0;
      if (!dump_string("}")) return 0;
      break;
    }
    case TP_U:
      if (!dump_source_name(TYPE_U_SOURCE_NAME(e))) return 0;
      if (!dump_string("{")) return 0;
      if (!dump_type(TYPE_U_TYPE(e))) return 0;
      if (!dump_string("}")) return 0;
      break;
  }

  return 1;
}

static int dump_function_args(int e, int has_return_type)
{
  int cur = e;
  if (has_return_type) cur = BARE_FUNCTION_TYPE_NEXT(cur);

  if (!cur) return 0;

  if (!dump_string("(")) return 0;
  while (cur) {
    if (!dump_type(BARE_FUNCTION_TYPE_TYPE(cur))) return 0;
    cur = BARE_FUNCTION_TYPE_NEXT(cur);
    if (cur) if (!dump_string(", ")) return 0;
  }
  if (!dump_string(")")) return 1;

  if (has_return_type) {
    if (!dump_string(" return type:")) return 0;
    if (!dump_type(BARE_FUNCTION_TYPE_TYPE(e))) return 0;
  }

  return 1;
}

static int dump_source_name(int e)
{
  return sputs(&out, names.data + SOURCE_NAME_STRING(e));
}

static int dump_operator_name(int e, int with_prefix)
{
  char *t;
  switch (OPERATOR_NAME_OP(e)) {
    default: return 0;
    case OP_UNARY_AND: t = "&"; break;
    case OP_AND: t = "&"; break;
    case OP_AFFECT: t = "="; break;
    case OP_AND_EQ: t = "&="; break;
    case OP_AND_AND: t = "&&"; break;
    case OP_NEG: t = "~"; break;
    case OP_COMMA: t = ","; break;
    case OP_CLASS: t = "()"; break;
    case OP_CAST:
      if (with_prefix) {
        if (!dump_string("operator cast{")) return 0;
      } else {
        if (!dump_string("cast{")) return 0;
      }
      if (!dump_type(OPERATOR_NAME_TYPE(e))) return 0;
      if (!dump_string("}")) return 0;
      return 1;
    case OP_DELETE: t = "delete"; break;
    case OP_DELETE_ARRAY: t = "delete[]"; break;
    case OP_POINTER: t = "*"; break;
    case OP_DIV: t = "/"; break;
    case OP_DIV_EQ: t = "/="; break;
    case OP_XOR: t = "^"; break;
    case OP_XOR_EQ: t = "^="; break;
    case OP_EQ: t = "=="; break;
    case OP_GT: t = ">"; break;
    case OP_GE: t = ">="; break;
    case OP_ARRAY: t = "[]"; break;
    case OP_LEFT_SHIFT: t = "<<"; break;
    case OP_LEFT_SHIFT_EQ: t = "<<="; break;
    case OP_LT: t = "<"; break;
    case OP_LE: t = "<="; break;
    case OP_MINUS: t = "-"; break;
    case OP_MUL: t = "*"; break;
    case OP_MINUS_EQ: t = "-="; break;
    case OP_MUL_EQ: t = "*="; break;
    case OP_DEC: t = "--"; break;
    case OP_NEW: t = "new"; break;
    case OP_NEW_ARRAY: t = "new[]"; break;
    case OP_MINUS_UNARY: t = "-"; break;
    case OP_NEQ: t = "!="; break;
    case OP_NOT: t = "!"; break;
    case OP_OR: t = "|"; break;
    case OP_OR_EQ: t = "|="; break;
    case OP_OR_OR: t = "||"; break;
    case OP_PLUS_UNARY: t = "+"; break;
    case OP_PLUS: t = "+"; break;
    case OP_PLUS_EQ: t = "+="; break;
    case OP_PLUS_PLUS: t = "++"; break;
    case OP_PM: t = "->*"; break;
    case OP_PT: t = "->"; break;
    case OP_QUESTION: t = "?"; break;
    case OP_MOD: t = "%"; break;
    case OP_MOD_EQ: t = "%="; break;
    case OP_RIGHT_SHIFT: t = ">>"; break;
    case OP_RIGHT_SHIFT_EQ: t = ">>="; break;
    case OP_SIZEOF: t = "sizeof"; break;
    case OP_VENDOR:
      if (with_prefix) if (!dump_string("operator ")) return 0;
      if (!dump_source_name(OPERATOR_NAME_SOURCE_NAME(e))) return 0;
      return 1;
  }
  if (with_prefix) {
    if (!dump_string("operator")) return 0;
    if (source[pos] >= 'a' && source[pos] <= 'z')
      if (!dump_string(" ")) return 0;
  }
  if (!dump_string(t)) return 0;
  return 1;
}

static int dump_unqualified_name(int e)
{
  switch (ELEM_TYPE(e)) {
    /* we can't have ctor/dtor here */
    default: return 0;
    case SOURCE_NAME:
      if (!dump_source_name(e)) return 0;
      break;
    case OPERATOR_NAME:
      if (!dump_operator_name(e, 1)) return 0;
      break;
  }
  return 1;
}

static int dump_ctor_dtor_name(int e, int elem)
{
  int last = 0;

  if (!CTOR_DTOR_NAME_CTOR(e)) if (!dump_string("~")) return 0;

  /* the constructor/destructor applies to a name, get the last
   * name of the nested name
   */
  while (elem) {
    int d = NESTED_NAME_DATA(elem);
    if (ELEM_TYPE(d) == SUBSTITUTION ||
        ELEM_TYPE(d) == SOURCE_NAME ||
        ELEM_TYPE(d) == OPERATOR_NAME)
      last = d;
    elem = NESTED_NAME_NEXT(elem);
  }
  if (!last) return 0;
  if (ELEM_TYPE(last) == SOURCE_NAME)
    return dump_source_name(last);
  else if (ELEM_TYPE(last) == OPERATOR_NAME)
    return dump_operator_name(last, 1);
  else switch (SUBSTITUTION_TYPE(last)) {
    case 't':
    case 'a':
    default: return dump_substitution(last);
    case 'b':
    case 's': return dump_string("basic_string");
    case 'i': return dump_string("basic_istream");
    case 'o': return dump_string("basic_ostream");
    case 'd': return dump_string("basic_iostream");
  }
}

static int dump_call_offset(int e)
{
  /* TODO: avoid string->int->string */
  char b[32];
  sprintf(b, "%d", CALL_OFFSET_1(e));
  if (!dump_string(b)) return 0;
  if (CALL_OFFSET_NB(e) == 1) return 1;
  sprintf(b, "%d", CALL_OFFSET_2(e));
  if (!dump_string(b)) return 0;
  return 1;
}

static int dump_encoding(int e)
{
  int n;
  int saved_templates = templates;
  templates = 0;

  n = ENCODING_NAME(e);

  /* look for the template */
  if (ELEM_TYPE(n) == NESTED_NAME) {
    int next;
    int cur;
    cur = n;
    next = NESTED_NAME_NEXT(n);
    while (cur && next) {
      cur = next;
      next = NESTED_NAME_NEXT(cur);
    }
    if (cur && ELEM_TYPE(NESTED_NAME_DATA(cur)) == TEMPLATE_ARGS)
      templates = NESTED_NAME_DATA(cur);
  } else if (ELEM_TYPE(n) == NAME) {
    if (NAME_TEMPLATE_ARGS(n))
      templates = NAME_TEMPLATE_ARGS(n);
  }

  switch(ELEM_TYPE(n)) {
    default: return 0;
    case NAME:
      if (NAME_ST(n)) if (!dump_string("std::")) return 0;
      if (ELEM_TYPE(NAME_NAME(n)) == SUBSTITUTION) {
        if (!dump_substitution(NAME_NAME(n))) return 0;
      } else
        if (!dump_unqualified_name(NAME_NAME(n))) return 0;
      if (NAME_TEMPLATE_ARGS(n)) {
        if (!dump_template_args(NAME_TEMPLATE_ARGS(n))) return 0;
        if (ENCODING_BARE_FUNCTION_TYPE(e))
          if (!dump_function_args(ENCODING_BARE_FUNCTION_TYPE(e), 1))
            return 0;
      } else {
        if (ENCODING_BARE_FUNCTION_TYPE(e))
          if (!dump_function_args(ENCODING_BARE_FUNCTION_TYPE(e), 0))
            return 0;
      }
      break;
    case NESTED_NAME: {
      int last_is_template_args = 0;
      int ctor_encountered = 0;
      int next;
      int cur;
      cur = n;
      next = NESTED_NAME_NEXT(n);
      while (cur) {
        int el = NESTED_NAME_DATA(cur);
        switch (ELEM_TYPE(el)) {
          default: return 0;
          case TYPE_TEMPLATE_PARAM:
            if (!dump_template_param(el)) return 0;
            last_is_template_args = 0;
            break;
          case SUBSTITUTION:
            if (!dump_substitution(el)) return 0;
            last_is_template_args = 0;
            break;
          case TEMPLATE_ARGS:
            if (!dump_template_args(el)) return 0;
            last_is_template_args = 1;
            break;
          case SOURCE_NAME:
            if (!dump_source_name(el)) return 0;
            last_is_template_args = 0;
            break;
          case CTOR_DTOR_NAME:
            if (!dump_ctor_dtor_name(el, n)) return 0;
            last_is_template_args = 0;
            ctor_encountered = 1;
            break;
          case OPERATOR_NAME:
            if (!dump_operator_name(el, 1)) return 0;
            last_is_template_args = 0;
            break;
        }
        if (next)
          if (ELEM_TYPE(NESTED_NAME_DATA(next)) != TEMPLATE_ARGS)
            if (!dump_string("::")) return 0;
        cur = next;
        if (next) next = NESTED_NAME_NEXT(cur);
      }
      if (ENCODING_BARE_FUNCTION_TYPE(e))
        if (!dump_function_args(ENCODING_BARE_FUNCTION_TYPE(e),
                                last_is_template_args && !ctor_encountered))
        return 0;
      if (NESTED_NAME_QUALIFIER_R(n))
        if (!dump_string(" restrict")) return 0;
      if (NESTED_NAME_QUALIFIER_V(n))
        if (!dump_string(" volatile")) return 0;
      if (NESTED_NAME_QUALIFIER_K(n))
        if (!dump_string(" const")) return 0;
      break;
    }
    case LOCAL_NAME:
      if (!dump_encoding(LOCAL_NAME_ENCODING(n))) return 0;
      if (!dump_string("::")) return 0;
      if (LOCAL_NAME_ENTITY(n)) {
        if (!dump_name(LOCAL_NAME_ENTITY(n))) return 0;
      } else {
        if (!dump_string("string literal")) return 0;
      }
      if (ENCODING_BARE_FUNCTION_TYPE(e))
        if (!dump_function_args(ENCODING_BARE_FUNCTION_TYPE(e), 0))
          return 0;
      break;
    case SPECIAL_NAME:
      switch (SPECIAL_NAME_TYPE(n)) {
        case 'V':
          if (!dump_string("virtual table{")) return 0;
          if (!dump_type(SPECIAL_NAME_TYPE_TYPE(n))) return 0;
          if (!dump_string("}")) return 0;
          break;
        case 'T':
          if (!dump_string("VTT{")) return 0;
          if (!dump_type(SPECIAL_NAME_TYPE_TYPE(n))) return 0;
          if (!dump_string("}")) return 0;
          break;
        case 'I':
          if (!dump_string("typeinfo{")) return 0;
          if (!dump_type(SPECIAL_NAME_TYPE_TYPE(n))) return 0;
          if (!dump_string("}")) return 0;
          break;
        case 'S':
          if (!dump_string("typeinfo name{")) return 0;
          if (!dump_type(SPECIAL_NAME_TYPE_TYPE(n))) return 0;
          if (!dump_string("}")) return 0;
          break;
        case 'G':
          if (!dump_string("guard variable{")) return 0;
          if (!dump_name(SPECIAL_NAME_NAME(n))) return 0;
          if (!dump_string("}")) return 0;
          break;
        case 'h':
        case 'v':
          if (!dump_string("thunk{")) return 0;
          if (!dump_encoding(SPECIAL_NAME_ENCODING(n))) return 0;
          if (!dump_string("}{")) return 0;
          if (!dump_call_offset(SPECIAL_NAME_CALL_OFFSET_1(n))) return 0;
          if (!dump_string("}")) return 0;
          break;
        case 'c':
          if (!dump_string("thunk{")) return 0;
          if (!dump_encoding(SPECIAL_NAME_ENCODING(n))) return 0;
          if (!dump_string("}{")) return 0;
          if (!dump_call_offset(SPECIAL_NAME_CALL_OFFSET_1(n))) return 0;
          if (!dump_string(", ")) return 0;
          if (!dump_call_offset(SPECIAL_NAME_CALL_OFFSET_2(n))) return 0;
          if (!dump_string("}")) return 0;
          break;
      }
      break;
  }

  templates = saved_templates;

  return 1;
}

static char *_demang(char *name)
{
  int e;
  source = name;
  pos = 0;
  elem_size = 0;
  names.size = 0;
  subs.size = 0;
  templates = 0;
  out.size = 0;

  if (!(e=mangled_name()) || source[pos]) return name;

//  if (!ENCODING_BARE_FUNCTION_TYPE(e)) return name;

  if (!dump_encoding(e)) return name;

  return out.data;
}

#else /* USE_CPP_DEMANGLER */

char *__cxa_demangle(const char *, char *, size_t *, int *);
/* this buffer is never freed, we should do it */
static char *demang_buf;
static size_t demang_size;

#endif /* USE_CPP_DEMANGLER */

/* if demangling fails (for it cannot be done or for there is
 * a bug in this code), name is returned
 */
char *demang(char *name)
{
#ifdef USE_CPP_DEMANGLER
  int status;
  char *ret;
  if (name[0] != '_' || name[1] != 'Z') return name;
  ret = __cxa_demangle(name, demang_buf, &demang_size, &status);
  if (!ret) return name;
  demang_buf = ret;
  return ret;
#else
  char *ret = _demang(name);
  if (ret == name && name[0] == '_' && name[1] == 'Z')
    fprintf(stderr, "DEMANGLER: cannot demangle %s\n", name);
  return ret;
#endif /* USE_CPP_DEMANGLER */
}

#if 0
int main(int n, char **v)
{
  int i;
  for (i = 1; i < n; i++) {
    char *t = demang(v[i]);
    if (t != v[i])
      fprintf(stdout, "%s: %s\n", v[i], t);
  }
  return 0;
}
#endif


/*
    <mangled-name> ::= _Z <encoding>

    <encoding> ::= <function name> <bare-function-type>
	       ::= <data name>
	       ::= <special-name>

    <name> ::= <nested-name>
	   ::= <unscoped-name>
	   ::= <unscoped-template-name> <template-args>
	   ::= <local-name>	# See Scope Encoding below
           ::= St <unqualified-name> # ::std::

    <unscoped-name> ::= <unqualified-name>
		    ::= St <unqualified-name>   # ::std::

    <unscoped-template-name> ::= <unscoped-name>
			     ::= <substitution>

    <nested-name> ::= N [<CV-qualifiers>] <prefix> <unqualified-name> E
		  ::= N [<CV-qualifiers>] <template-prefix> <template-args> E

    <prefix> ::= <prefix> <unqualified-name>
	     ::= <template-prefix> <template-args>
             ::= <template-param>
	     ::= # empty
	     ::= <substitution>

    <template-prefix> ::= <prefix> <template unqualified-name>
                      ::= <template-param>
                      ::= <substitution>

    <unqualified-name> ::= <operator-name>
                       ::= <ctor-dtor-name>  
                       ::= <source-name>   

    <source-name> ::= <positive length number> <identifier>

    <number> ::= [n] <non-negative decimal integer>

    <identifier> ::= <unqualified source code identifier>


    <operator-name> ::= nw	# new           
		  ::= na	# new[]
		  ::= dl	# delete        
		  ::= da	# delete[]      
		  ::= ps        # + (unary)
		  ::= ng	# - (unary)     
		  ::= ad	# & (unary)     
		  ::= de	# * (unary)     
		  ::= co	# ~             
		  ::= pl	# +             
		  ::= mi	# -             
		  ::= ml	# *             
		  ::= dv	# /             
		  ::= rm	# %             
		  ::= an	# &             
		  ::= or	# |             
		  ::= eo	# ^             
		  ::= aS	# =             
		  ::= pL	# +=            
		  ::= mI	# -=            
		  ::= mL	# *=            
		  ::= dV	# /=            
		  ::= rM	# %=            
		  ::= aN	# &=            
		  ::= oR	# |=            
		  ::= eO	# ^=            
		  ::= ls	# <<            
		  ::= rs	# >>            
		  ::= lS	# <<=           
		  ::= rS	# >>=           
		  ::= eq	# ==            
		  ::= ne	# !=            
		  ::= lt	# <             
		  ::= gt	# >             
		  ::= le	# <=            
		  ::= ge	# >=            
		  ::= nt	# !             
		  ::= aa	# &&            
		  ::= oo	# ||            
		  ::= pp	# ++            
		  ::= mm	# --            
		  ::= cm	# ,             
		  ::= pm	# ->*           
		  ::= pt	# ->            
		  ::= cl	# ()            
		  ::= ix	# []            
		  ::= qu	# ?             
		  ::= st	# sizeof (a type)
		  ::= sz	# sizeof (an expression)
		  ::= cv <type>	# (cast)        
		  ::= v <digit> <source-name>	# vendor extended operator

  <special-name> ::= TV <type>	# virtual table
		 ::= TT <type>	# VTT structure (construction vtable index)
		 ::= TI <type>	# typeinfo structure
		 ::= TS <type>	# typeinfo name (null-terminated byte string)
                 ::= GV <object name>	# Guard variable for one-time initialization
			# No <type>
                 ::= T <call-offset> <base encoding>
		      # base is the nominal target function of thunk
                 ::= Tc <call-offset> <call-offset> <base encoding>
		      # base is the nominal target function of thunk
		      # first call-offset is 'this' adjustment
		      # second call-offset is result adjustment

  <call-offset> ::= h <nv-offset> _
		::= v <v-offset> _

  <nv-offset> ::= <offset number>
		      # non-virtual base override

  <v-offset>  ::= <offset number> _ <virtual offset number>
		      # virtual base override, with vcall offset


  <ctor-dtor-name> ::= C1	# complete object constructor
		   ::= C2	# base object constructor
		   ::= C3	# complete object allocating constructor
		   ::= D0	# deleting destructor
		   ::= D1	# complete object destructor
		   ::= D2	# base object destructor

  <type> ::= <builtin-type>
	 ::= <function-type>
	 ::= <class-enum-type>
	 ::= <array-type>
	 ::= <pointer-to-member-type>
	 ::= <template-param>
	 ::= <template-template-param> <template-args>
	 ::= <substitution> # See Compression below

  <type> ::= <CV-qualifiers> <type>
	 ::= P <type>	# pointer-to
	 ::= R <type>	# reference-to
         ::= O <type>     # rvalue reference-to (C++0x)
	 ::= C <type>	# complex pair (C 2000)
	 ::= G <type>	# imaginary (C 2000)
	 ::= U <source-name> <type>	# vendor extended type qualifier

  <CV-qualifiers> ::= [r] [V] [K] 	# restrict (C99), volatile, const

  <builtin-type> ::= v	# void
		 ::= w	# wchar_t
		 ::= b	# bool
		 ::= c	# char
		 ::= a	# signed char
		 ::= h	# unsigned char
		 ::= s	# short
		 ::= t	# unsigned short
		 ::= i	# int
		 ::= j	# unsigned int
		 ::= l	# long
		 ::= m	# unsigned long
		 ::= x	# long long, __int64
		 ::= y	# unsigned long long, __int64
		 ::= n	# __int128
		 ::= o	# unsigned __int128
		 ::= f	# float
		 ::= d	# double
		 ::= e	# long double, __float80
		 ::= g	# __float128
		 ::= z	# ellipsis
		 ::= u <source-name>	# vendor extended type

  <function-type> ::= F [Y] <bare-function-type> E

  <bare-function-type> ::= <signature type>+
	# types are possible return type, then parameter types

  <class-enum-type> ::= <name>

  <array-type> ::= A <positive dimension number> _ <element type>
	       ::= A [<dimension expression>] _ <element type>

  <pointer-to-member-type> ::= M <class type> <member type>

  <template-param> ::= T_	# first template parameter
		   ::= T <parameter-2 non-negative number> _

  <template-template-param> ::= <template-param>
			    ::= <substitution>

  <template-args> ::= I <template-arg>+ E

  <template-arg> ::= <type>			# type or template
		 ::= X <expression> E		# expression
                 ::= <expr-primary>             # simple expressions

  <expression> ::= <unary operator-name> <expression>
	       ::= <binary operator-name> <expression> <expression>
	       ::= <trinary operator-name> <expression> <expression> <expression>
               ::= st <type>
               ::= <template-param>
               ::= sr <type> <unqualified-name>                   # dependent name
               ::= sr <type> <unqualified-name> <template-args>   # dependent template-id
	       ::= <expr-primary>

  <expr-primary> ::= L <type> <value number> E                   # integer literal
                 ::= L <type <value float> E                     # floating literal
                 ::= L <mangled-name> E                           # external name

  <local-name> := Z <function encoding> E <entity name> [<discriminator>]
               := Z <function encoding> E s [<discriminator>]

  <discriminator> := _ <non-negative number> 

  <substitution> ::= S <seq-id> _
		 ::= S_

   <substitution> ::= St # ::std::
   <substitution> ::= Sa # ::std::allocator
   <substitution> ::= Sb # ::std::basic_string
   <substitution> ::= Ss # ::std::basic_string < char,
						 ::std::char_traits<char>,
						 ::std::allocator<char> >
   <substitution> ::= Si # ::std::basic_istream<char,  std::char_traits<char> >
   <substitution> ::= So # ::std::basic_ostream<char,  std::char_traits<char> >
   <substitution> ::= Sd # ::std::basic_iostream<char, std::char_traits<char> >
*/
