#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

#include "symlist.h"
#include "call_graph.h"
#include "hist.h"

#include "gui.h"
#include "elf.h"

static FILE *gmon;

int da_read(unsigned long *v)
{
  int r;
  unsigned long c;
  if (big_endian) {
    if ((r=getc(gmon)) == EOF) return -1;
    c = (r&255)<<24;
    if ((r=getc(gmon)) == EOF) return -1;
    c |= (r&255)<<16;
    if ((r=getc(gmon)) == EOF) return -1;
    c |= (r&255)<<24;
    if ((r=getc(gmon)) == EOF) return -1;
    c |= r&255;
  } else {
    if ((r=getc(gmon)) == EOF) return -1;
    c = r;
    if ((r=getc(gmon)) == EOF) return -1;
    c |= (r&255)<<8;
    if ((r=getc(gmon)) == EOF) return -1;
    c |= (r&255)<<16;
    if ((r=getc(gmon)) == EOF) return -1;
    c |= (r&255)<<24;
  }
  *v = c;
  return 0;
}

int da_read64(uint64_t *v)
{
  int r;
  uint64_t c;
  uint64_t cur;
  if (big_endian) {
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255; cur <<= 32+24;
    c = cur;
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255; cur <<= 32+16;
    c |= cur;
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255; cur <<= 32+8;
    c |= cur;
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255; cur <<= 32;
    c |= cur;
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255; cur <<= 24;
    c |= cur;
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255; cur <<= 16;
    c |= cur;
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255; cur <<= 8;
    c |= cur;
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255;
    c |= cur;
  } else {
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255;
    c = cur;
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255; cur <<= 8;
    c |= cur;
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255; cur <<= 16;
    c |= cur;
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255; cur <<= 24;
    c |= cur;
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255; cur <<= 32;
    c |= cur;
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255; cur <<= 32+8;
    c |= cur;
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255; cur <<= 32+16;
    c |= cur;
    if ((r=getc(gmon)) == EOF) return -1;
    cur = r&255; cur <<= 32+24;
    c |= cur;
  }
  *v = c;
  return 0;
}

#define DOread \
  do { \
    if (da_read(&type)==-1) {fprintf(stderr, "read error\n"); exit(1); } \
  } while(0)

#define DOreadPTR \
  do { \
    if (bit_mode == 32) { \
      if (da_read(&type)==-1) {fprintf(stderr, "read error\n"); exit(1); } \
    } else { \
      uint64_t x; \
      if (da_read64(&x)==-1) {fprintf(stderr, "read error\n"); exit(1); } \
      if (sizeof(unsigned long) < 8) { \
        if ((x&((uint64_t)(0xfffffffflu))) != x) { \
          fprintf(stderr, "64b/32b stuff, contact sed@free.fr for a fix\n"); \
          exit(1); \
        } \
      } \
      type = x; \
    } \
  } while(0)

#define DOreadn(from, n) \
  do { \
    int w; \
    /* fprintf(stderr, "reading %d", n); */ \
    for (w=0; w<n; w++) if (getc(gmon) == EOF) { \
      fprintf(stderr, "read n (=%d) (from=%d) error\n", n, from); \
      exit(1); \
    } \
  } while(0)

#define DOread2(x) \
  do { \
    int c; \
    if (big_endian) { \
      c = getc(gmon); if(c==EOF){fprintf(stderr, "read2 error 1\n");exit(1);}\
      x = 0;\
      x |= (c&255)<<8;\
      c = getc(gmon); if(c==EOF){fprintf(stderr, "read2 error 2\n");exit(1);}\
      x |= (c&255);\
    } else { \
      c = getc(gmon); if(c==EOF){fprintf(stderr, "read2 error 3\n");exit(1);}\
      x = 0;\
      x |= (c&255);\
      c = getc(gmon); if(c==EOF){fprintf(stderr, "read2 error 4\n");exit(1);}\
      x |= ((c&255)<<8);\
   } \
  } while(0)

#if MY_MAIN

int main(int argn, char **argv)
{
  hist h;
  int nb_hist = 0;
  int nb_arc = 0;
  unsigned long type;
  unsigned long caller, called, count;
  int n;
  int i;
  gui g;

  h.size = 0;

  if (argn != 3) {
    fprintf(stderr, "usage: xgprof <ELF program> <gmon file>\n");
    return 1;
  }

  if (add_null_symbol() == -1) return 1;

  fprintf(stderr, "sucking symbols from program...\n");

  if (get_elf_symbols(argv[1]) == -1) return 1;

  fprintf(stderr, "... done\n");

  fprintf(stderr, "%d symbols found\n", nb_symbols());

  fprintf(stderr, "sorting symbols...\n");

  if (end_add_symbol() == -1) {
    fprintf(stderr, "oom in end_add_symbol?\n");
    exit(1);
  }

  fprintf(stderr, "%d unique symbols found\n", nb_unique_symbols());

  fprintf(stderr, "... done\n");

  fprintf(stderr, "building call graph...\n");

  gmon = fopen(argv[2], "r");
  if (!gmon) {
    perror(argv[2]);
    return 1;
  }

  /* header */
  DOread;
  DOread;
  DOread;
  DOread;
  DOread;
  /* blocks */
  while (1) {
    fprintf(stderr, "."); fflush(stderr);
    type = getc(gmon); if (type == EOF) break;
    switch(type) {
      case 0: /* hist */ /* hist_read_rec */ {
        unsigned long lowpc, highpc;
        unsigned long val;
        nb_hist++;
//fprintf(stderr, "hist\n");
        /* read header */
        DOreadPTR; /* record->lowpc */
        lowpc = type;
fprintf(stderr, "histogram lowpc=%8.8lx\n", lowpc);
        DOreadPTR; /* record->highpc */
        highpc = type;
fprintf(stderr, "histogram highpc=%8.8lx\n", highpc);
        DOread; /* record->num_bins */
        n = type;
fprintf(stderr, "histogram nb bins=%d\n", n);
        DOread; /* profrate */
        if (type != 100) {
          fprintf(stderr, "rate of hist = %lu (Hz I guess)\n", type);
          fprintf(stderr, "100 is expected, ciao.\n");
          exit(1);
        }
        if (init_hist(&h, lowpc, highpc, n) == -1) {
          fprintf(stderr, "error in init_hist\n");
          exit(1);
        }
        DOreadn(-1, 15); /* char n_hist_dimension[15] */
        DOreadn(-2, 1); /* char n_hist_dimension_abbrev */
        /* read stuff */
        for (i=0; i<n; i++) {
          DOread2(val);
          add_hist(&h, i, val);
        }
        break;
      }
      case 1: /* arc */ /* cg_read_rec */
        nb_arc++;
//fprintf(stderr, "arc\n");
        DOreadPTR; /* from_pc */
        caller = type;
//fprintf(stderr, "from ok %p\n", (void*)type);
        DOreadPTR; /* self_pc */
        called = type;
//fprintf(stderr, "self ok %p\n", (void*)type);
        DOread; /* count */
        count = type;
//fprintf(stderr, "arc count %d\n", type);
        if (add_call_item(caller, called, count) == -1) {
          fprintf(stderr, "oom in add_call_item?\n");
          exit(1);
        }
        break;
      case 2: /* bb_count */ /* bb_read_rec */
        fprintf(stderr, "bb_count not supported yet\n"); exit(1);
        break;
      default: fprintf(stderr, "bad type %lu\n", type); exit(1);
    }
  }
  fprintf(stderr, "\n%d hist %d arc found\n", nb_hist, nb_arc);

  fclose(gmon);

  if (end_items() == -1) {
    fprintf(stderr, "oom in end_items?\n");
    exit(1);
  }

  fprintf(stderr, "... done\n");

  fprintf(stderr, "handling histogram data...\n");

  if (calculate_hist(&h) == -1) {
    fprintf(stderr, "trouble with calculate_hist\n");
    exit(1);
  }

  fprintf(stderr, "... done\n");

  fprintf(stderr, "sorting symbols...\n");

  if (sort_symbols() == -1) {
    fprintf(stderr, "trouble with sort_symbols\n");
    exit(1);
  }

  fprintf(stderr, "... done\n");

  dump_calls();

  fprintf(stderr, "switching to GUI\n");

  gtk_init(&argn, &argv);

  if (init_gui(&g) == -1) {
    fprintf(stderr, "problem initing gui\n");
    return 1;
  }
  gtk_widget_show_all(g.win);

  gtk_main();

  return 0;
}

#endif
